# MIT License
#
# Copyright (c) 2021 willfuks
#
# Permission is hereby granted, free of charge, to any person obtaining a copy
# of this software and associated documentation files (the "Software"), to deal
# in the Software without restriction, including without limitation the rights
# to use, copy, modify, merge, publish, distribute, sublicense, and/or sell
# copies of the Software, and to permit persons to whom the Software is
# furnished to do so, subject to the following conditions:
#
# The above copyright notice and this permission notice shall be included in all
# copies or substantial portions of the Software.
#
# THE SOFTWARE IS PROVIDED "AS IS", WITHOUT WARRANTY OF ANY KIND, EXPRESS OR
# IMPLIED, INCLUDING BUT NOT LIMITED TO THE WARRANTIES OF MERCHANTABILITY,
# FITNESS FOR A PARTICULAR PURPOSE AND NONINFRINGEMENT. IN NO EVENT SHALL THE
# AUTHORS OR COPYRIGHT HOLDERS BE LIABLE FOR ANY CLAIM, DAMAGES OR OTHER
# LIABILITY, WHETHER IN AN ACTION OF CONTRACT, TORT OR OTHERWISE, ARISING FROM,
# OUT OF OR IN CONNECTION WITH THE SOFTWARE OR THE USE OR OTHER DEALINGS IN THE
# SOFTWARE.


from django.db import models


class JWTBlacklist(models.Model):
    jti = models.CharField(max_length=255, db_index=True, unique=True, null=False,
                           blank=False)
    token = models.TextField(db_index=True, null=True)
    expires = models.DateTimeField(db_index=True, null=True)
    created_at = models.DateTimeField(auto_now_add=True)

    def __str__(self):  # pragma: no cover
        created_at_str = (
            self.created_at.strftime("%Y-%m-%d %H:%M:%S") if self.created_at else None
        )
        expires_str = (
            self.expires.strftime("%Y-%m-%d %H:%M:%S") if self.expires else None
        )
        return (
            f'Token jti is: {self.jti}, {"Created at " if created_at_str else None}'
            f'{created_at_str}'
            f'{" and expires at " if expires_str else None}{expires_str}.'
        )

    @staticmethod
    def is_blacklisted(jti: str):
        """
        Checks if token is already blacklisted.

        Arguments
        ---------
          jti: str
              JWT token format.

        Returns
        -------
          bool
              If `True` then token already exists on db.

        Raises
        ------
        """
        return JWTBlacklist.objects.filter(jti=jti).exists()



# MIT License
#
# Copyright (c) 2021 willfuks
#
# Permission is hereby granted, free of charge, to any person obtaining a copy
# of this software and associated documentation files (the "Software"), to deal
# in the Software without restriction, including without limitation the rights
# to use, copy, modify, merge, publish, distribute, sublicense, and/or sell
# copies of the Software, and to permit persons to whom the Software is
# furnished to do so, subject to the following conditions:
#
# The above copyright notice and this permission notice shall be included in all
# copies or substantial portions of the Software.
#
# THE SOFTWARE IS PROVIDED "AS IS", WITHOUT WARRANTY OF ANY KIND, EXPRESS OR
# IMPLIED, INCLUDING BUT NOT LIMITED TO THE WARRANTIES OF MERCHANTABILITY,
# FITNESS FOR A PARTICULAR PURPOSE AND NONINFRINGEMENT. IN NO EVENT SHALL THE
# AUTHORS OR COPYRIGHT HOLDERS BE LIABLE FOR ANY CLAIM, DAMAGES OR OTHER
# LIABILITY, WHETHER IN AN ACTION OF CONTRACT, TORT OR OTHERWISE, ARISING FROM,
# OUT OF OR IN CONNECTION WITH THE SOFTWARE OR THE USE OR OTHER DEALINGS IN THE
# SOFTWARE.


from __future__ import annotations

import base64
import json
from calendar import timegm
from datetime import datetime
from typing import Any, Callable, Dict

from django.contrib.auth import get_user_model
from django.core.exceptions import ImproperlyConfigured, ValidationError
from django.core.serializers.json import DjangoJSONEncoder
from django.http.request import HttpRequest
from django.http.response import HttpResponseBase, JsonResponse
from django.middleware.csrf import rotate_token
from django.utils.decorators import method_decorator
from django.views import View
from django.views.decorators.csrf import (csrf_exempt, csrf_protect,
                                          ensure_csrf_cookie)

import djwto.authentication as auth
import djwto.models as models
import djwto.settings as settings
import djwto.signals as signals
import djwto.tokens as tokens
from djwto.exceptions import JWTValidationError

HEADERS401 = {'WWW-Authenticate': auth.WWWAUTHENTICATE}


def identity(func: Callable) -> Callable:
    """
    Helper function to be used when `settings.CHECK_CSRF` is `False`; in this case the
    dispatch decorators should operate just as normally and therefore we return the
    function (or method) itself as if nothing really changed.

    Arguments
    ---------
      func: Callable
          Function to return directly. Usually this will be `post` methods defined in
          views.

    Returns
    -------
      func: Callable
          Same function as input, nothing changes when CSRF is not enabled.
    """
    return func


def _build_decorator(func: Callable) -> Callable:
    """
    If CSRF is enabled and there are cookies being saved then return the input `func`
    wrapped by the `method_decorator` function.

    Arguments
    ---------
      func: Callable
          Function to wrap and make it available for methods decorator. Usually will be
          either `ensure_csrf_cookie` or `csrf_protect`.

    Returns
    -------
      func: Callable
          Function decorated to operate with methods.
    """
    if settings.DJWTO_CSRF and settings.DJWTO_MODE in {'ONE-COOKIE', 'TWO-COOKIES'}:
        return method_decorator(func)
    return method_decorator(identity)


def build_tokens_response(
    refresh_claims: Dict[str, Any],
    access_claims: Dict[str, Any],
    msg: str = None
) -> JsonResponse:
    """
    djwto offers three main modes for returning a response: "JSON", "ONE-COOKIE" or
    "TWO-COOKIES". This method builds the http response in accordance to what
    `settings.DJWTO_MODE` specifies.

    For "JSON" option, the tokens as simply put into a dictionary and returned as a
    JSON response type. For "ONE-COOKIE", each token (refresh and access) is saved
    into a separate respective cookie.

    In "TWO-COOKIES" mode, the access token is divided in two partes:
    - "jwt_access_payload": contains the serialized value of the claims stored in an
    open Http Cookie so the front-end can read it.
    - "jwt_access_token": jwt token value, stored in a protected cookie for safety.

    The payload part is publicly available for reading by the client whereas the
    signature is saved as HttpOnly. This allows for the front-end to interact with
    the JWT content without running the risk of compromising the signature.

    Args
    ----
      refresh_claims: Dict[str, Any]
          Token already encoded.
      access_claims: Dict[str, Any]
          If `settings.DJWTO_MODE == 'TWO-COOKIES'` then it's expected the value of
          `access_claims` will be serialized in the cookie content. This allows for
          the front-end to have access to its values. The jwt token cookie is still
          separated and stored under `HttpOnly` and `Secure` conditions.
      msg: str
          Message to return in JSON response to the client.

    Returns
    -------
      JsonResponse
          Returns tokens in accordance to `settings.DJWTO_MODE` value.
    """
    refresh_token = tokens.encode_claims(refresh_claims)
    access_token = tokens.encode_claims(access_claims)

    mode = settings.DJWTO_MODE
    if mode == 'JSON':
        return JsonResponse({'refresh': refresh_token, 'access': access_token})

    refresh_lifetime = settings.DJWTO_REFRESH_TOKEN_LIFETIME
    max_age_refresh = (
        int(refresh_lifetime.total_seconds()) if refresh_lifetime else None
    )

    access_lifetime = settings.DJWTO_ACCESS_TOKEN_LIFETIME
    max_age_access = (
        int(access_lifetime.total_seconds()) if access_lifetime else None
    )

    response = JsonResponse({'msg': msg})
    response.set_cookie(
        'jwt_refresh',
        refresh_token,
        max_age=max_age_refresh,
        httponly=True,
        secure=True,
        path=f'/{settings.DJWTO_REFRESH_COOKIE_PATH}',
        samesite=settings.DJWTO_SAME_SITE,
        domain=settings.DJWTO_DOMAIN
    )
    if mode == 'ONE-COOKIE':
        response.set_cookie(
            'jwt_access',
            access_token,
            max_age=max_age_access,
            httponly=True,
            secure=True,
            samesite=settings.DJWTO_SAME_SITE,
            domain=settings.DJWTO_DOMAIN
        )
        return response
    if mode == 'TWO-COOKIES':
        response.set_cookie(
            'jwt_access_payload',
            base64.b64encode(
                json.dumps(access_claims, sort_keys=True,
                           cls=DjangoJSONEncoder).encode(),
            ).decode(),
            max_age=max_age_access,
            httponly=False,
            secure=True,
            samesite=settings.DJWTO_SAME_SITE,
            domain=settings.DJWTO_DOMAIN
        )
        response.set_cookie(
            'jwt_access_token',
            access_token,
            max_age=max_age_access,
            httponly=True,
            secure=True,
            samesite=settings.DJWTO_SAME_SITE,
            domain=settings.DJWTO_DOMAIN
        )
        return response
    raise ImproperlyConfigured(
        'settings.DJWTO_MODE must be either "JSON", "ONE-COOKIE" or "TWO-COOKIES".'
        f'Received "{mode}" instead.'
    )


class RefreshAccessView(View):
    """
    Uses the refresh token to create a new access one.
    """
    @method_decorator(csrf_exempt)
    def dispatch(
        self, request: HttpRequest, *args: Any, **kwargs: Any
    ) -> HttpResponseBase:
        return super().dispatch(request, *args, **kwargs)

    @_build_decorator(csrf_protect)
    @method_decorator(auth.jwt_login_required)
    @method_decorator(auth.jwt_is_refresh)
    def post(
        self,
        request: auth.THttpRequest,
        *args: Any,
        **kwargs: Any
    ) -> JsonResponse:
        refresh_claims = request.payload

        if settings.DJWTO_JTI_CLAIM:
            jti = refresh_claims['jti']
            if models.JWTBlacklist.is_blacklisted(jti):
                return JsonResponse({'error': "Can't update access token."},
                                    status=500)

        access_claims = tokens.get_access_claims_from_refresh(refresh_claims)
        msg = 'Access token successfully refreshed.'
        response = build_tokens_response(refresh_claims, access_claims, msg)

        signals.jwt_access_refreshed.send(
            sender=type(self).__name__,
            request=request,
            refresh_claims=refresh_claims,
            access_claims=access_claims
        )

        return response


class UpdateRefreshView(View):
    """
    Updates the expiration time of the refresh token. This is particularly useful in
    scenarios such as web eCommerces where the customer might be close to finalize its
    purchase and the token goes expired during the checkout.

    In this view the client has the opportunity of updating the refresh token thus
    avoiding a login requirement on inconvenient moments.

    The `settings.DJWTO_ALLOW_REFRESH_UPDATE` must be `True`, the token must not be
    blacklisted already (valid only if "JTI" is available) and the user must have a
    `is_active` flag equal to `True` from the db (if "user" is available in JWT token).
    """
    @method_decorator(csrf_exempt)
    def dispatch(
        self, request: HttpRequest, *args: Any, **kwargs: Any
    ) -> HttpResponseBase:
        return super().dispatch(request, *args, **kwargs)

    @_build_decorator(csrf_protect)
    @method_decorator(auth.jwt_login_required)
    @method_decorator(auth.jwt_is_refresh)
    def post(
        self,
        request: auth.THttpRequest,
        *args: Any,
        **kwargs: Any
    ) -> JsonResponse:
        fail_resp = JsonResponse({'error': "Can't update refresh token."}, status=500)
        refresh_claims = request.payload

        if not settings.DJWTO_ALLOW_REFRESH_UPDATE:
            return fail_resp

        if settings.DJWTO_JTI_CLAIM:
            jti = refresh_claims['jti']
            if models.JWTBlacklist.is_blacklisted(jti):
                return fail_resp

        user_data = refresh_claims.get('user')
        if user_data:
            User = get_user_model()
            try:
                username = user_data.get(User.USERNAME_FIELD)
                # TODO: typing is not recognizing `get_by_natural_key` here
                user = User.objects.get_by_natural_key(username)  # type: ignore
            except User.DoesNotExist:
                return fail_resp
            if not user.is_active:
                return JsonResponse({'error': 'User is inactive.'}, status=403)

        iat = datetime.utcnow()
        refresh_claims['exp'] = timegm(
            (iat + settings.DJWTO_REFRESH_TOKEN_LIFETIME).utctimetuple()
        )
        if 'iat' in refresh_claims:
            refresh_claims['iat'] = timegm(iat.utctimetuple())
        access_claims = tokens.get_access_claims_from_refresh(refresh_claims)
        msg = 'Refresh token successfully updated.'
        response = build_tokens_response(refresh_claims, access_claims, msg)

        signals.jwt_refresh_updated.send(
            sender=type(self).__name__,
            request=request,
            refresh_claims=refresh_claims,
            access_claims=access_claims
        )

        return response


class GetTokensView(View):
    """
    Creates the JWT Token and stores then according to the specification in
    `settings.DJWTO_MODE`.
    """
    @method_decorator(csrf_exempt)
    def dispatch(
        self, request: HttpRequest, *args: Any, **kwargs: Any
    ) -> HttpResponseBase:
        return super().dispatch(request, *args, **kwargs)

    @_build_decorator(ensure_csrf_cookie)
    def post(
        self,
        request: auth.THttpRequest,
        *args: Any,
        **kwargs: Any
    ) -> JsonResponse:
        try:
            _ = auth.get_raw_token_from_request(request)
            return JsonResponse({'msg': 'User already authenticated.'}, status=200)
        except (JWTValidationError, ImproperlyConfigured):
            pass

        try:
            user = auth.user_authenticate(request)
        except ValidationError as err:

            signals.jwt_login_fail.send(
                sender=type(self).__name__,
                request=request,
                error=err.args[0]
            )

            return JsonResponse({'error': err.args[0]}, status=401,
                                headers=HEADERS401)

        try:
            refresh_claims = tokens.process_claims(user, request, 'refresh', args,
                                                   kwargs)
            access_claims = tokens.get_access_claims_from_refresh(refresh_claims)
            msg = 'Tokens successfully created.'
            response = build_tokens_response(refresh_claims,  access_claims, msg)

            signals.jwt_logged_in.send(
                sender=type(self).__name__,
                request=request,
                refresh_claims=refresh_claims,
                access_claims=access_claims
            )

            rotate_token(request)

            return response

        except ImproperlyConfigured as err:

            signals.jwt_login_fail.send(
                sender=type(self).__name__,
                request=request,
                error=err.args[0]
            )
            # As it's an internal error then don't return full error message to the
            # client. It should be handled internally by the server side.
            return JsonResponse({'error': 'Failed to process request.'}, status=500)


class ValidateTokensView(View):
    """
    Extracts the jwt token from income `request` and assess if the token is valid.
    """
    @method_decorator(csrf_exempt)
    def dispatch(
        self, request: HttpRequest, *args: Any, **kwargs: Any
    ) -> HttpResponseBase:
        return super().dispatch(request, *args, **kwargs)

    @_build_decorator(csrf_protect)
    @method_decorator(auth.jwt_login_required)
    def post(
        self,
        request: auth.THttpRequest,
        *args: Any,
        **kwargs: Any
    ) -> JsonResponse:
        """
        In some scenarios it's interesting to have an endpoint which a client can ask
        whether a given token is valid or not. Notice that even if the token has been
        Blacklisted but it's still valid the response will confirm its validity.

        It's been designed like that because blacklisted tokens are still valid for its
        expiration time. Only when requesting for updating the refresh token is that the
        blacklist API is checked.
        """
        signals.jwt_token_validated.send(sender=type(self).__name__, request=request)
        return JsonResponse({'msg': 'Token is valid'})


class BlackListTokenView(View):
    """
    Extracts the jwt token from income `request` and adds it to the Blacklist table.
    The Blacklist functionality is only available if `settings.DJWTO_JTI_CLAIM` is `True`

    Only the refresh token can be blacklisted. After the operation, available cookies are
    deleted accordingly as well.
    """
    @method_decorator(csrf_exempt)
    def dispatch(
        self, request: HttpRequest, *args: Any, **kwargs: Any
    ) -> HttpResponseBase:
        return super().dispatch(request, *args, **kwargs)

    @_build_decorator(csrf_protect)
    @method_decorator(auth.jwt_login_required)
    @method_decorator(auth.jwt_is_refresh)
    def post(
        self,
        request: auth.THttpRequest,
        *args: Any,
        **kwargs: Any
    ) -> JsonResponse:
        jti = request.payload.get('jti')
        if not jti:
            error_msg = (
                'No jti claim was available in the input token. The value is mandatory'
                ' in order to use the Blacklist API.'
            )
            return JsonResponse({'error': error_msg}, status=403)

        # jti key is checked in case another token was created with the same jti but
        # different expire time (this should not happen so this checking works as a
        # protection against that.
        if models.JWTBlacklist.is_blacklisted(jti):
            error_msg = (
                'Input jti token is already blacklisted.'
            )
            return JsonResponse({'error': error_msg}, status=409)

        exp = request.payload.get('exp')
        if exp:
            exp = datetime.utcfromtimestamp(exp)

        models.JWTBlacklist.objects.get_or_create(
            jti=jti,
            token=request.token,
            expires=exp
        )

        signals.jwt_blacklisted.send(
            sender=type(self).__name__,
            request=request,
            jti=jti
        )

        return JsonResponse({'msg': 'Token successfully blacklisted.'}, status=200)

    @_build_decorator(csrf_protect)
    @method_decorator(auth.jwt_login_required)
    def delete(
        self,
        request: auth.THttpRequest,
        *args: Any,
        **kwargs: Any
    ) -> JsonResponse:
        """
        Deletes both tokens cookies.
        """
        if settings.DJWTO_MODE == 'JSON':
            return JsonResponse({'msg': 'No token to delete.'}, status=204)

        response = JsonResponse({'msg': 'Tokens successfully deleted.'})
        response.delete_cookie(
            'jwt_refresh',
            path=f'/{settings.DJWTO_REFRESH_COOKIE_PATH}',
            domain=settings.DJWTO_DOMAIN,
            samesite=settings.DJWTO_SAME_SITE
        )

        if settings.DJWTO_MODE == 'ONE-COOKIE':
            response.delete_cookie(
                'jwt_access',
                domain=settings.DJWTO_DOMAIN,
                samesite=settings.DJWTO_SAME_SITE
            )

        if settings.DJWTO_MODE == 'TWO-COOKIES':
            response.delete_cookie(
                'jwt_access_payload',
                domain=settings.DJWTO_DOMAIN,
                samesite=settings.DJWTO_SAME_SITE
            )
            response.delete_cookie(
                'jwt_access_token',
                domain=settings.DJWTO_DOMAIN,
                samesite=settings.DJWTO_SAME_SITE
            )
        return response



# MIT License
#
# Copyright (c) 2021 willfuks
#
# Permission is hereby granted, free of charge, to any person obtaining a copy
# of this software and associated documentation files (the "Software"), to deal
# in the Software without restriction, including without limitation the rights
# to use, copy, modify, merge, publish, distribute, sublicense, and/or sell
# copies of the Software, and to permit persons to whom the Software is
# furnished to do so, subject to the following conditions:
#
# The above copyright notice and this permission notice shall be included in all
# copies or substantial portions of the Software.
#
# THE SOFTWARE IS PROVIDED "AS IS", WITHOUT WARRANTY OF ANY KIND, EXPRESS OR
# IMPLIED, INCLUDING BUT NOT LIMITED TO THE WARRANTIES OF MERCHANTABILITY,
# FITNESS FOR A PARTICULAR PURPOSE AND NONINFRINGEMENT. IN NO EVENT SHALL THE
# AUTHORS OR COPYRIGHT HOLDERS BE LIABLE FOR ANY CLAIM, DAMAGES OR OTHER
# LIABILITY, WHETHER IN AN ACTION OF CONTRACT, TORT OR OTHERWISE, ARISING FROM,
# OUT OF OR IN CONNECTION WITH THE SOFTWARE OR THE USE OR OTHER DEALINGS IN THE
# SOFTWARE.


import json
from functools import wraps
from typing import Any, Callable, Dict, List, Tuple, cast

from django.contrib.auth.forms import AuthenticationForm
from django.contrib.auth.models import User
from django.core.exceptions import ImproperlyConfigured, ValidationError
from django.http.request import HttpRequest
from django.http.response import JsonResponse

import djwto.settings as settings
import djwto.tokens as tokens
from djwto.exceptions import JWTValidationError


class THttpRequest(HttpRequest):
    """
    Helper class to workaround mypy not recognizing attributes `token` and `payload`
    in `HttpRequest`.
    """
    payload: Dict[str, Any]
    token: str


WWWAUTHENTICATE = 'Bearer realm=api'


def jwt_passes_test(
    test_func: Callable[[THttpRequest], Tuple[bool, str]]
) -> Callable:
    """
    Function based on the original Django's `user_passes_test` but adapted to work with
    the jwt tokens instead.

    Args
    ----
      test_func: Callable[[Dict[str, Any]], bool]
          Function that returns `True` if tests succeeds and `False` otherwise. It
          receives the payload of the jwt token as input.

    Returns
    -------
      Callable
          Decorator function to receive the input `view_func` and process the request.
    """
    def decorator(view_func: Callable) -> Callable:
        @wraps(view_func)
        def _wrapped_view(
            request: THttpRequest, *args: Any, **kwargs: Any
        ) -> JsonResponse:
            worked, error = test_func(request)
            if error:
                return JsonResponse({'error': error}, status=403)
            if worked:
                return view_func(request, *args, **kwargs)
            else:
                return JsonResponse({'error': 'Failed to validate token.'}, status=403)
        return _wrapped_view
    return decorator


def jwt_login_required(func: Callable) -> Callable:
    """
    Similar to the original Django's `login_required` but functioning on top of the jwt
    token as processed from the request. This function specifically enforces on views
    the requirement for the tokens been already created and validated in order for the
    dispatch to continue.

    Args
    ----
      func: Callable
          Refers to one of the verbs a view can have, such as `POST` or `GET`.

    Returns
    -------
      Callable
          Decorator that returns function to process view if and only if the requried
          jwt tokens for authentication are available in the income `request` object.
    """
    def test_func(request: THttpRequest) -> Tuple[bool, str]:
        try:
            token = get_raw_token_from_request(request)
            payload = tokens.decode_token(token)
            request.payload = payload
            request.token = token
            return ('username' and 'id') in payload.get('user', {}), ''
        except (JWTValidationError, ImproperlyConfigured) as err:
            return False, str(err)
    return jwt_passes_test(test_func)(func)


def jwt_perm_required(perms: List[str]) -> Callable:
    """
    Receives a list of permissions and only process the inner function if the jwt token
    available in the input request have the required permissions.
    """
    def process_func(func: Callable) -> Callable:
        """
        Similar to the original Django's `permission_required` but functioning on top of
        the jwt token as processed from the request. This function specifically enforces
        on views that the input token has a field inside of the claim `user` called
        `perms` containing the specified permission to process the input function.

        Args
        ----
          func: Callable
              Refers to one of the verbs a view can have, such as `POST` or `GET`.

        Returns
        -------
          Callable
              Decorator that returns function to process view if and only if the requried
              jwt tokens contains the necessary permissions.
        """
        def test_func(request: THttpRequest) -> Tuple[bool, str]:
            # It's supposed by default that `jwt_login_required` comes first. This
            # guarantees that `request.payload` will not be `None`
            if not getattr(request, 'payload', ''):
                return False, 'Login must happen before evaluating permissions.'
            payload = request.payload
            jwt_perms = payload.get('user', {}).get('perms')
            if jwt_perms:
                if not all([perm in jwt_perms for perm in perms]):
                    return False, 'Insufficient Permissions.'
                return True, ''
            return False, 'Invalid permissions for jwt token.'
        return jwt_passes_test(test_func)(func)
    return process_func


def jwt_is_refresh(view_func: Callable) -> Callable:
    """
    Some views only work with the refresh token. This decorator validates whether the
    input request has a refresh token or not, returning error response if the token is
    not found.

    Args
    ----
      view_func: Callable
          Refers to one of the verbs a view can have, such as `POST` or `GET`.

    Returns
    -------
      Callable
          Decorator that returns function to process view if and only if the requried
          jwt tokens for authentication are available in the income `request` object.
    """
    def test_func(request: THttpRequest) -> Tuple[bool, str]:
        # In this case token should be in Header AUTHORIZATION
        if settings.DJWTO_MODE == 'JSON':
            if request.payload.get('type') != 'refresh':
                return False, 'Refresh token was not sent in authorization header.'
            return True, ''
        path = settings.DJWTO_REFRESH_COOKIE_PATH
        if path not in request.path:
            return False, f'Refresh token is only sent in path: {path}'
        if request.method == 'POST' and request.POST.get('jwt_type') != 'refresh':
            return False, 'POST variable "jwt_type" must be equal to "refresh".'
        return True, ''
    return jwt_passes_test(test_func)(view_func)


def user_authenticate(request: HttpRequest) -> User:
    """
    Default method used for authenticating users; uses authentication layer as offered by
    Django.

    Args
    ----
      request: HttpRequest
          Request object as received by the web server processor (either WSGI or ASGI),
          expected to contain the field `data` filled with user name and password.

    Returns
    -------
      User
          If user successfully authenticate then returns their correspondent object from
          Django's default `User` model.

    Raises
    ------
      ValidationError: If validation of `form` fails.
    """
    form = AuthenticationForm(data=request.POST)
    form.is_valid()
    if form.errors:
        raise ValidationError(json.dumps(dict(form.errors)))
    # cast to `User` type as `get_user` returns AbstractUser
    user = cast(User, form.get_user())
    return user


def jwt_authenticate(request: HttpRequest) -> Dict[str, Any]:
    """
    Extract token expected to be sent in input `request` and assess its validity. If
    still valid, returns general information available in the payload. Notice that
    in djwto package no request to the database is performed, that is, the entire
    validation process is based solely on whether the input token is a valid one or
    not. This approach follows precisely what JWT tokens were designed for.

    Args
    ----
      request: HttpRequest
          Input request processed from WSGI (or ASGI) server.

    Returns
    -------
      validated_token: Dict[str, Any]
          JWT token after validation process.

    Raises
    ------
      JWTValidationError: If input token is invalid.
    """
    token = get_raw_token_from_request(request)
    validated_token = tokens.decode_token(token)
    return validated_token


def get_raw_token_from_request(request: HttpRequest) -> str:
    """
    Returns string as encoded by pyJWT library from request. Its storage can be in
    three different places according to `settings.DJWTO_MODE`:
      - 'JSON': the token is expected to be found in the `HTTP_AUTHORIZATION` header
        with value like "Authorization: Bearer abc.def.ghi".
      - 'ONE-COOKIE': the access token is expected to be found in a cookie named
        'jwt_access'.
      - 'TWO-COOKIES': the token is divided into two cookies, one called
        'jwt_access_payload' which contains the header and payload of the token and
        'jwt_access_signature' containing the signature to validate the payload.

    `request` may contain a data field named 'jwt_type' which is either 'access'
    (default) or 'refresh'. This value sets which token cookie to retrieve and is
    only valid if mode type also sets cookies.

    Arguments
    ---------
      request: HttpRequest
          Input request from WSGI (or ASGI) server.

    Returns
    -------
      str
          JWT token in the format abc.def.ghi

    Raises
    ------
      JWTValidationError:
          If token is not in header.
          If token is invalid.
          If token is empty.
          If cookies are not available.
      ImproperlyConfigured:
          If `settings.DJWTO_MODE` has invalid value.
    """
    if settings.DJWTO_MODE == 'JSON':
        auth_header = request.META.get('HTTP_AUTHORIZATION', '')
        if not auth_header:
            raise JWTValidationError(
                'Token not found in "HTTP_AUTHORIZATION" header.'
            )
        token = auth_header.split('Bearer ')[-1]
        # Either token or the string 'Bearer ' are not available
        if not token or len(token) == len(auth_header):
            raise JWTValidationError(
                'Value in HTTP_AUTHORIZATION header is not valid.'
            )
        return token

    # Defaults to 'access' to make it easier on the front-end calls
    type_ = request.POST.get('jwt_type', 'access')
    if type_ not in {'access', 'refresh'}:
        raise JWTValidationError(
            'Input data "jwt_type" must be either "access" or "refresh".'
            f' Got "{type_}" instead.'
        )

    if settings.DJWTO_MODE == 'ONE-COOKIE' or type_ == 'refresh':
        token = request.COOKIES.get(f'jwt_{type_}', '')
        if not token:
            if (
                type_ == 'refresh' and
                settings.DJWTO_REFRESH_COOKIE_PATH not in request.path
            ):
                raise JWTValidationError(
                    'Refresh cookie is only sent in path '
                    f'"/{settings.DJWTO_REFRESH_COOKIE_PATH}". Requested path was: '
                    f'{request.path}.'
                )
            raise JWTValidationError(f'Cookie "jwt_{type_}" cannot be empty.')
        return token

    # If it made until here then only option is to retrieve access token
    if settings.DJWTO_MODE == 'TWO-COOKIES':
        token = request.COOKIES.get('jwt_access_token')
        if not token:
            raise JWTValidationError(
                'Cookie "jwt_access_token" cannot be empty.'
            )
        return token
    raise ImproperlyConfigured(
        'Value of `settings.DJWTO_MODE` is invalid. Expected either "JSON", '
        f'"ONE-COOKIE" or "TWO-COOKIES". Received "{settings.DJWTO_MODE}" instead.'
    )



from django.db import migrations, models


class Migration(migrations.Migration):

    initial = True

    dependencies = [
    ]

    operations = [
        migrations.CreateModel(
            name='JWTBlacklist',
            fields=[
                ('id', models.AutoField(auto_created=True, primary_key=True, serialize=False, verbose_name='ID')),
                ('jti', models.CharField(db_index=True, max_length=255, unique=True)),
                ('token', models.TextField(db_index=True, null=True)),
                ('expires', models.DateTimeField(db_index=True, null=True)),
                ('created_at', models.DateTimeField(auto_now_add=True)),
            ],
        ),
    ]



# MIT License
#
# Copyright (c) 2021 willfuks
#
# Permission is hereby granted, free of charge, to any person obtaining a copy
# of this software and associated documentation files (the "Software"), to deal
# in the Software without restriction, including without limitation the rights
# to use, copy, modify, merge, publish, distribute, sublicense, and/or sell
# copies of the Software, and to permit persons to whom the Software is
# furnished to do so, subject to the following conditions:
#
# The above copyright notice and this permission notice shall be included in all
# copies or substantial portions of the Software.
#
# THE SOFTWARE IS PROVIDED "AS IS", WITHOUT WARRANTY OF ANY KIND, EXPRESS OR
# IMPLIED, INCLUDING BUT NOT LIMITED TO THE WARRANTIES OF MERCHANTABILITY,
# FITNESS FOR A PARTICULAR PURPOSE AND NONINFRINGEMENT. IN NO EVENT SHALL THE
# AUTHORS OR COPYRIGHT HOLDERS BE LIABLE FOR ANY CLAIM, DAMAGES OR OTHER
# LIABILITY, WHETHER IN AN ACTION OF CONTRACT, TORT OR OTHERWISE, ARISING FROM,
# OUT OF OR IN CONNECTION WITH THE SOFTWARE OR THE USE OR OTHER DEALINGS IN THE
# SOFTWARE.


import uuid
from calendar import timegm
from datetime import datetime, timedelta
from typing import Any, Dict, List, Optional, Union

import jwt as pyjwt
from django.contrib.auth.models import User
from django.core.exceptions import ImproperlyConfigured
from django.http import HttpRequest

import djwto.exceptions as exceptions
import djwto.settings as settings


def process_claims(
    user: Optional[User],
    request: HttpRequest,
    type_: str = 'refresh',
    *args: Any,
    **kwargs: Any
) -> Dict[Any, Any]:
    """
    Default processing of claims that are usually available on JWTs such as `sub`, `iss`,
    `aud` and so on. It's intention is to work as a general solution for most use cases
    of JWT creation.

    Separating this step in a contained function allows for customizing the values used
    on each JWT claim, such as, for instance, extracting a specific value of 'audience'
    for each user based on Cookies or Meta data available in input `request`.

    The default version simply uses values available on the file `settings.py`.

    Args
    ----
      user: Optional[User]
          Object as retrieved from the database representing the user. It's optional so
          extensions of this function doesn't have to input a user object case it's not
          required.
      request: HttpRequest
          Input request received from the WSGI (or ASGI) server.
      type_: str
          Whether these claims refers to a refresh or access token.
      args, kwargs: Any
          Those remain as a signature reference for the scenario where one wants to
          extend the functionality of this function.

    Returns
    ------
      claims: Dict[Any, Any]
          Default values of claim fields as specified in the `settings.py` file.

    Raises
    ------
      ImproperlyConfigured: If values in file `settings.py` are invalid.
    """
    claims: Dict[Any, Any] = {}
    iss = settings.DJWTO_ISS_CLAIM
    if iss:
        if not isinstance(iss, str):
            raise ImproperlyConfigured(
                f'Value of Issuer must be str. Received {type(iss).__name__} instead.'
            )
        claims['iss'] = iss

    sub = settings.DJWTO_SUB_CLAIM
    if sub:
        if not isinstance(sub, str):
            raise ImproperlyConfigured(
                f'Value of Subject must be str. Received {type(sub).__name__} instead.'
            )
        claims['sub'] = sub

    aud = settings.DJWTO_AUD_CLAIM
    if aud:
        if not isinstance(aud, list):
            if not isinstance(aud, str):
                raise ImproperlyConfigured(
                    'Value of Audience must be either List[str] or str. '
                    f'Received {type(aud).__name__} instead.'
                )
        else:
            for e in aud:
                if not isinstance(e, str):
                    raise ImproperlyConfigured(
                        'Value of Audience must be either List[str] or str. '
                        f'Received List[{type(aud).__name__}] instead.'
                    )
        claims['aud'] = aud

    # First it builds claims for refresh token and then access token is copied from it.
    exp_timedelta = settings.DJWTO_REFRESH_TOKEN_LIFETIME

    iat = datetime.utcnow()
    if settings.DJWTO_IAT_CLAIM:
        claims['iat'] = timegm(iat.utctimetuple())

    if exp_timedelta:
        _validate_timedelta_claim(exp_timedelta)
        exp = iat + exp_timedelta
        claims['exp'] = timegm(exp.utctimetuple())

    nbf_timedelta = settings.DJWTO_NBF_LIFETIME
    if nbf_timedelta:
        _validate_timedelta_claim(exp_timedelta)
        nbf = iat + nbf_timedelta
        claims['nbf'] = timegm(nbf.utctimetuple())

    if settings.DJWTO_JTI_CLAIM:
        jti = str(uuid.uuid4())
        claims['jti'] = jti

    # This is necessary for the case when mode is 'JSON' as it's the only way to
    # identify which token is which
    claims['type'] = 'refresh'

    if user:
        claims['user'] = process_user(user)
    return claims


def process_user(user: User) -> Dict[str, Union[str, int, List[str]]]:
    """
    Process various data related to the database User object. This function can be
    replaced in order to build a different claim set for the user.

    Args
    ----
      user: User
          Object user as retrieved from the database.

    Returns
    -------
      Dict[str, str]
          User claims to store in the token payload.
    """
    return {
        user.USERNAME_FIELD: user.get_username(),
        'id': user.pk,
        'perms': process_perms(user)
    }


def process_perms(user: User) -> List[str]:
    """
    Gets permissions from input user and returns a list with those values. Default
    function gets all permissions (by running `user.get_all_permissions()`). Change this
    funtion, by running:

    >>> import djwto.authentication as auth
    >>> auth.process_perms = new_perms_func

    Args
    ----
      user: User
          Input user as retrieved from database.

    Returns
    ------
      perms: List[str]
          List of permissions extracted from user. By default returns all perms.
    """
    return list(user.get_all_permissions())


def _validate_timedelta_claim(claim: Optional[timedelta]) -> None:
    """
    Receives as input the setting value for a given claim that may contain timedelta
    type, checking if it's valid.

    Args
    ----
      claim: Optional[timedelta]
          Value from input setting.

    Raises
    ------
      ImproperlyConfigured: If claim is not of type `timedelta`.
                  If timdelta is negative.
    """
    if not isinstance(claim, timedelta):
        raise ImproperlyConfigured(
            'Refresh token lifetime must be a `timedelta` object.'
        )
    if claim.total_seconds() < 0:
        raise ImproperlyConfigured(
            'Refresh token expiration time must be positive.'
        )


def get_access_claims_from_refresh(refresh_claims: Dict[Any, Any]) -> Dict[Any, Any]:
    """
    Uses the claims built for the refresh token to create the access one. Main difference
    is that the latter might have a lower expiration date (still, this field is optional,
    if `settings.DJWTO_ACCESS_TOKEN_LIFETIME` is `None` then access token won't have it.

    Args
    ----
      refresh_claims: Dict[Any, Any]
          Dict object where keys are the claims names of the refresh token as extracted
          from user and `request` input.

    Returns
    -------
      access_claims: Dict[Any, Any]
          It's basically a copy of the refresh token with a few changes, if applicable.

    Raises
    ------
      ImproperlyConfigured: If claim is not of type `timedelta`.
                  If timdelta is negative.
    """
    # Uses a shallow copy for now. Maybe in the future it might require a deep copy.
    # This forces the tokens to have the claims at the first level of the dict.
    access_claims = refresh_claims.copy()
    iat = datetime.utcnow()
    if settings.DJWTO_IAT_CLAIM:
        # Keep the value of the emission of the refresh token which allows for the front-
        # end to read its value on `DJWTO_MODE == 'TWO-COOKIES'` option and evaluate
        # whether it's necessary to create a new refresh token. This is useful for
        # instance on eCommerce websites that might not want to logout a customer that is
        # still currently active on the website.
        refresh_iat = access_claims['iat']
        access_claims['iat'] = timegm(iat.utctimetuple())
        access_claims['refresh_iat'] = refresh_iat
    access_timedelta = settings.DJWTO_ACCESS_TOKEN_LIFETIME
    access_claims['type'] = 'access'
    if access_timedelta:
        _validate_timedelta_claim(access_timedelta)
        access_claims['exp'] = timegm((iat + access_timedelta).utctimetuple())
    return access_claims


def encode_claims(claims: Dict[Any, Any]) -> str:
    """
    Uses the dict of claims built from the request and transforms to JWT using the chosen
    algorithm and key from the `settings.py` file.

    Args
    ----
      claims: Dict[Any, Any]
          Input claims to be encoded in JWT.

    Returns
    -------
      str
          Final encoded JWT form.
    """
    return pyjwt.encode(claims, settings.DJWTO_SIGNING_KEY,
                        algorithm=settings.DJWTO_ALGORITHM)


def decode_token(token: str) -> Dict[str, Any]:
    """
    Validates if input token (in the form of 'abc.def.ghi') is a valid JWT token. The
    token is considered valid if the HMAC of the payload is equal to the signature,
    as well as the expiration time is still bigger than current date or the field
    'nbf' is lower. If either 'exp' nor 'nbf' are available then skip those tests.

    Arguments
    ---------
      token: str
          JWT as extracted from request.

    Returns
    -------
      Tuple[Dict[Any, Any]
          If validated, returns the token payload in dict format.


    Args
    ----
      token: str
    """
    sign_key = settings.DJWTO_SIGNING_KEY
    ver_key = settings.DJWTO_VERIFYING_KEY
    alg = [settings.DJWTO_ALGORITHM]
    iss = settings.DJWTO_ISS_CLAIM
    sub = settings.DJWTO_SUB_CLAIM
    aud = settings.DJWTO_AUD_CLAIM
    iat_flag = settings.DJWTO_IAT_CLAIM
    jti_flag = settings.DJWTO_JTI_CLAIM

    kwargs: Dict[str, Any] = {}
    require: List[str] = []

    def _update(name: str, value: Any) -> None:
        if value:
            kwargs[name] = value
            require.append(name[:3])

    _update('issuer', iss)
    _update('subject', sub)
    _update('audience', aud)
    _update('iat', iat_flag)
    _update('jti', jti_flag)

    kwargs['options'] = {'require': require}

    try:
        payload = pyjwt.decode(token, ver_key if ver_key else sign_key, alg, **kwargs)
        return payload
    except (
        pyjwt.ExpiredSignatureError,
        pyjwt.ImmatureSignatureError,
        pyjwt.MissingRequiredClaimError,
        pyjwt.DecodeError,
        pyjwt.InvalidTokenError
    ) as err:
        raise exceptions.JWTValidationError(str(err))



# MIT License
#
# Copyright (c) 2021 willfuks
#
# Permission is hereby granted, free of charge, to any person obtaining a copy
# of this software and associated documentation files (the "Software"), to deal
# in the Software without restriction, including without limitation the rights
# to use, copy, modify, merge, publish, distribute, sublicense, and/or sell
# copies of the Software, and to permit persons to whom the Software is
# furnished to do so, subject to the following conditions:
#
# The above copyright notice and this permission notice shall be included in all
# copies or substantial portions of the Software.
#
# THE SOFTWARE IS PROVIDED "AS IS", WITHOUT WARRANTY OF ANY KIND, EXPRESS OR
# IMPLIED, INCLUDING BUT NOT LIMITED TO THE WARRANTIES OF MERCHANTABILITY,
# FITNESS FOR A PARTICULAR PURPOSE AND NONINFRINGEMENT. IN NO EVENT SHALL THE
# AUTHORS OR COPYRIGHT HOLDERS BE LIABLE FOR ANY CLAIM, DAMAGES OR OTHER
# LIABILITY, WHETHER IN AN ACTION OF CONTRACT, TORT OR OTHERWISE, ARISING FROM,
# OUT OF OR IN CONNECTION WITH THE SOFTWARE OR THE USE OR OTHER DEALINGS IN THE
# SOFTWARE.

# type: ignore

import os
import sys

from codecs import open
from setuptools import setup, find_packages


here = os.path.abspath(os.path.dirname(__file__))

install_requires = [
    'Django',
    'pyjwt[crypto]',
    'typing_extensions'
]
tests_require = [
    'tox'
]
setup_requires = [
    'flake8',
    'mypy',
]

packages = find_packages()

_version = {}
_version_path = os.path.join(here, 'djwto', '__version__.py')

with open(_version_path, 'r', 'utf-8') as f:
    exec(f.read(), _version)

with open('README.md', 'r', 'utf-8') as f:
    readme = f.read()

setup(
    name='djwto',
    version=_version['__version__'],
    author='Willian Fuks',
    author_email='willian.fuks@gmail.com',
    description= "JWT based Auth layer built on top of Django.",
    long_description=readme,
    long_description_content_type='text/markdown',
    packages=packages,
    include_package_data=True,
    install_requires=install_requires,
    tests_require=tests_require,
    setup_requires=setup_requires,
    license='MIT License',
    classifiers=[
        'Development Status :: 3 - Alpha',
        'Environment :: Console',
        'Framework :: Django',
        'Intended Audience :: Developers',
        'Intended Audience :: Education',
        'Intended Audience :: End Users/Desktop',
        'License :: OSI Approved :: MIT License',
        'Natural Language :: English',
        'Operating System :: Unix',
        'Operating System :: Microsoft :: Windows',
        'Topic :: Internet',
        'Topic :: Security',
        'Programming Language :: Python :: 3.7',
        'Programming Language :: Python :: 3.8',
        'Programming Language :: Python :: 3.9',
        'Programming Language :: Python :: Implementation :: CPython',
        'Typing :: Typed'
    ],
    python_requires='>=3',
)



# MIT License
#
# Copyright (c) 2021 willfuks
#
# Permission is hereby granted, free of charge, to any person obtaining a copy
# of this software and associated documentation files (the "Software"), to deal
# in the Software without restriction, including without limitation the rights
# to use, copy, modify, merge, publish, distribute, sublicense, and/or sell
# copies of the Software, and to permit persons to whom the Software is
# furnished to do so, subject to the following conditions:
#
# The above copyright notice and this permission notice shall be included in all
# copies or substantial portions of the Software.
#
# THE SOFTWARE IS PROVIDED "AS IS", WITHOUT WARRANTY OF ANY KIND, EXPRESS OR
# IMPLIED, INCLUDING BUT NOT LIMITED TO THE WARRANTIES OF MERCHANTABILITY,
# FITNESS FOR A PARTICULAR PURPOSE AND NONINFRINGEMENT. IN NO EVENT SHALL THE
# AUTHORS OR COPYRIGHT HOLDERS BE LIABLE FOR ANY CLAIM, DAMAGES OR OTHER
# LIABILITY, WHETHER IN AN ACTION OF CONTRACT, TORT OR OTHERWISE, ARISING FROM,
# OUT OF OR IN CONNECTION WITH THE SOFTWARE OR THE USE OR OTHER DEALINGS IN THE
# SOFTWARE.

import os
from datetime import timedelta
from typing import List, Optional, Union

from django.conf import settings
from typing_extensions import Literal

# Refence for each claim: https://tools.ietf.org/html/rfc7519#page-9
DJWTO_ISS_CLAIM: Optional[str] = getattr(settings, 'DJWTO_ISS_CLAIM', None)
DJWTO_SUB_CLAIM: Optional[str] = getattr(settings, 'DJWTO_SUB_CLAIM', None)
DJWTO_AUD_CLAIM: Optional[Union[List[str], str]] = getattr(settings, 'DJWTO_AUD_CLAIM',
                                                           None)

DJWTO_IAT_CLAIM: bool = getattr(settings, 'DJWTO_IAT_CLAIM', True)
DJWTO_JTI_CLAIM: bool = getattr(settings, 'DJWTO_JTI_CLAIM', True)

DJWTO_ALLOW_REFRESH_UPDATE: bool = getattr(settings, 'DJWTO_ALLOW_REFRESH_UPDATE', True)

DJWTO_ACCESS_TOKEN_LIFETIME = getattr(settings, 'DJWTO_ACCESS_TOKEN_LIFETIME',
                                      timedelta(minutes=5))
DJWTO_REFRESH_TOKEN_LIFETIME = getattr(settings, 'DJWTO_REFRESH_TOKEN_LIFETIME',
                                       timedelta(days=1))
DJWTO_NBF_LIFETIME: Optional[timedelta] = getattr(settings, 'DJWTO_NBF_LIFETIME',
                                                  timedelta(minutes=0))

DJWTO_SIGNING_KEY: str = getattr(settings, 'DJWTO_SIGNING_KEY',
                                 os.environ['DJWTO_SIGNING_KEY'])

# Only set if Algorithm uses asymetrical signing.
DJWTO_VERIFYING_KEY: Optional[str] = getattr(settings, 'DJWTO_VERIFYING_KEY', None)

DJWTO_ALGORITHM: str = getattr(settings, 'DJWTO_ALGORITHM', 'HS256')

DJWTO_MODE: Literal['JSON', 'ONE-COOKIE', 'TWO-COOKIES'] = getattr(settings,
                                                                   'DJWTO_MODE', 'JSON')

DJWTO_REFRESH_COOKIE_PATH: str = getattr(settings, 'DJWTO_REFRESH_COOKIE_PATH',
                                                   'api/token/refresh')

DJWTO_SAME_SITE: str = getattr(settings, 'DJWTO_SAME_SITE', 'Lax')
DJWTO_DOMAIN: Optional[str] = getattr(settings, 'DJWTO_DOMAIN', None)
DJWTO_CSRF: bool = getattr(settings, 'DJWTO_CSRF', True)
